#!/usr/bin/env python

"""
highlighter.py

KPart plugin for KHTML to highlight text in an HTML page which matches a
given keyword.

Copyright (C) 2003-5 David Boddie <david@boddie.org.uk>

This software is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as
published by the Free Software Foundation; either version 2 of
the License, or (at your option) any later version.

This software is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public
License along with this library; see the file COPYING
If not, write to the Free Software Foundation, Inc.,
59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

"""

# Import the required Qt and KDE classes.

from qt import QCheckBox, QColor, QComboBox, QHBox, QLabel, QMessageBox, \
               QString, Qt, QWidget, QVBoxLayout, PYSIGNAL, SIGNAL

from kdecore import i18n, KAccel, KInstance, KLibFactory, KShortcut, KURL
from kparts import KParts
from kdeui import KAction, KColorButton, KLineEdit, KPushButton
from khtml import KHTMLPart, DOM


# Define a class which will be initialised when the plugin is loaded.

class FactoryClass(KLibFactory):

    instance = None
    
    def __init__(self, parent = None, name = None):
    
        #print "FactoryClass.__init__%s" % repr((parent, name))
        
        KLibFactory.__init__(self, parent, name)
        
        if FactoryClass.instance is None:
        
            FactoryClass.instance = KInstance("Highlighter")
    
    def createObject(self, parent = None, name = None, class_name = "QObject",
                     args = None):
    
        if args is None:
        
            args = []
        
        #print "FactoryClass.createObject%s" % repr((parent, name, class_name, args))
        
        plugin = Highlighter(parent, name)
        #print "created %s" % repr(plugin)
        
        # Either of these appear to work:
        #self.emit(SIGNAL("objectCreated(KParts::Plugin)"), (self.plugin,))
        self.emit(SIGNAL("objectCreated(QObject)"), (plugin,))
        
        #print "returning"
        return plugin


# Define the plugin class, derived from the standard KParts.Plugin class.

class Highlighter(KParts.Plugin):

    def __init__(self, parent = None, name = None):
    
        KParts.Plugin.__init__(self, parent, name)
        
        self._action = KAction(
            "Highlight keywords",
            "highlighter",
            KShortcut("Ctrl+H"),
            self.slotOpenKeywordDialog,
            self.actionCollection(),
            "highlightwords"
            )
        
        self.dialog = SearchDialog()
        self.connect(self.dialog, PYSIGNAL("findWord"), self.highlight)
    
    def slotOpenKeywordDialog(self):
    
        # The parent is assumed to be a KHTMLPart
        
        khtmlpart = self.parent()

        if not khtmlpart.inherits("KHTMLPart"):
        
            title = i18n( "Cannot highlight text" )
            text = i18n( "You can only highlight text in web pages." )
            
            QMessageBox.warning( None, title, text )
            return
        
        # Open a window to use to specify the search keyword/pattern.
        self.dialog.resize(self.dialog.minimumSize())
        self.dialog.show()
    
    def highlight(self):
    
        # The parent is assumed to be a KHTMLPart
        
        if self.dialog.text() == u"":
        
            title = i18n( "Cannot highlight text" )
            text = i18n( "You must specify a text string." )
            
            QMessageBox.warning( None, title, text )
            return
        
        # Find matching text in the HTML page and modify the DOM structure
        # to highlight it.
        
        dom = self.parent().document()
        
        root = dom.documentElement()
        
        frames = self.find_frames(root)
        
        if not frames:
        
            title = i18n( "Cannot highlight text" )
            text = i18n( "Failed to find a document body." )
            
            QMessageBox.warning( None, title, text )
            return
        
        # Search for text now because this won't work after the document
        # has been modified.
        self.parent().findTextBegin()
        self.parent().findTextNext(
            self.dialog.text(), self.dialog.forwards(), self.dialog.case(),
            False
            )
        
        for frame in frames:
        
            self.find_matching(
                dom, frame, self.dialog.text(), self.dialog.case(),
                self.dialog.colours()
                )
    
    def find_frames(self, node, nodes = None):
    
        if nodes is None:
        
            nodes = []
        
        name, value, attributes = self.node_info(node)
        
        if name.upper() == u"BODY":
        
            nodes.append(node)
            return nodes
        
        #elif name.upper() == u"FRAME":
        #
        #    nodes.append(node)
        #    return nodes
        
        # Examine any child nodes.
        
        child = node.firstChild()
        
        while not child.isNull():
        
            nodes = nodes + self.find_frames(child)
            
            child = child.nextSibling()
        
        return nodes
    
    def find_matching(self, dom, node, text, case, colours):
    
        # Examine any child nodes.
        
        child = node.firstChild()
        
        while not child.isNull():
        
            # Retrieve the child node from the matching method.
            
            # This is performed in case the child node is replaced by
            # a set of nodes for highlighted text. Otherwise, the original
            # child node may have been removed from the document and we
            # would lose track of our position.
            child = self.find_matching(dom, child, text, case, colours)
            
            child = child.nextSibling()
        
        if node.nodeType() != DOM.Node.TEXT_NODE:
        
            # Not a text node or it contains no text.
            
            # Return the existing node to the calling function.
            return node
        
        name, value, attributes = self.node_info(node)
        
        # Look for matching text.
        begin_at = self.match(value, text, 0, case)
        
        if begin_at != -1:
        
            # Matching text was found.
            
            # Replace this node with a collection in which the text given
            # is highlighted using <span>...</span> elements.
            
            span = dom.createElement( DOM.DOMString("span") )
            
            if begin_at == 0 and len(value) == len(text):
            
                # The node matches exactly.
                
                highlighted_span = dom.createElement( DOM.DOMString("span") )
                highlighted_span.setAttribute(
                    DOM.DOMString(u"style"),
                    DOM.DOMString(
                        u"color: %s; background-color: %s" % colours
                        )
                    )
                
                text_node = dom.createTextNode( DOM.DOMString(value) )
                self.set_attributes(text_node, attributes)
                
                highlighted_span.appendChild(text_node)
                
                span.appendChild(highlighted_span)
            
            else:
            
                # Split the node into pieces, if necessary.
                
                # Start looking for matches from the beginning of the
                # text node's value to allow preceding non-matching text
                # to be incorporated into the new set of nodes.
                
                last_end = 0
                
                while last_end < len(value):
                
                    begin = self.match(value, text, last_end, case)
                    
                    if begin > last_end:
                    
                        # Include the text up to the next match.
                        
                        text_node = dom.createTextNode(
                            DOM.DOMString(value[last_end:begin])
                            )
                        
                        self.set_attributes(text_node, attributes)
                        span.appendChild(text_node)
                    
                    elif begin == -1:
                    
                        # No more instances found.
                        
                        # Include the rest of the text found and exit the
                        # loop.
                        
                        text_node = dom.createTextNode(
                            DOM.DOMString(value[last_end:])
                            )
                        
                        self.set_attributes(text_node, attributes)
                        span.appendChild(text_node)
                        
                        break
                    
                    # Wrap the highlighted keyword in a <span>...</span> pair.
                    
                    highlighted_span = dom.createElement(
                        DOM.DOMString("span")
                        )
                    highlighted_span.setAttribute(
                        DOM.DOMString("style"),
                        DOM.DOMString(
                            u"color: %s; background-color: %s" % colours
                            )
                        )
                    
                    end = begin + len(text)
                    
                    text_node = dom.createTextNode(
                        DOM.DOMString(value[begin:end])
                        )
                    
                    self.set_attributes(text_node, attributes)
                    highlighted_span.appendChild(text_node)
                    
                    span.appendChild(highlighted_span)
                    
                    # Record the end of the match found, ready for the next
                    # search.
                    last_end = end
            
            # Replace the current node.
            parent = node.parentNode()
            parent.replaceChild(span, node)
            
            # Return the new set of nodes, contained within the span
            # element to the calling function.
            return span
        
        # Return the existing node to the calling function.
        return node
    
    def match(self, value, text, start, case):
    
        if case:
        
            return value.find(text, start)
        
        else:
        
            return value.upper().find(text.upper(), start)
    
    def node_info(self, node):
    
        name = unicode(node.nodeName().string())
        value = unicode(node.nodeValue().string())
        
        attrs = node.attributes()
        attributes = {}
        
        for a in range(0, attrs.length()):
        
            item = attrs.item(a)
            attr_name = unicode(item.nodeName().string()).lower()
            attributes[attr_name] = unicode(item.nodeValue().string())
        
        return name, value, attributes
    
    def set_attributes(self, node, attributes):
    
        for name, value in attributes.items():
        
            node.setAttribute( DOM.DOMString(name), DOM.DOMString(value) )


class SearchDialog(QWidget):

    def __init__(self):
    
        QWidget.__init__(self)
        
        self.setCaption( i18n("Highlighter") )
        
        spacing = 3
        
        # Create a vertical layout.
        layout = QVBoxLayout(self, spacing, spacing)
        
        # Create a label.
        label = QLabel(self)
        label.setText(
            i18n("Search pattern")
            )
        
        # Create an editable line of text.
        self.line_edit = KLineEdit(self)
        
        # Add checkboxes to control the direction and case sensitivity of
        # the search.
        self.direction_combo = QComboBox(self)
        self.direction_combo.insertItem(i18n("Search forwards"))
        self.direction_combo.insertItem(i18n("Search backwards"))
        
        self.case_checkbox = QCheckBox(
            i18n("C&ase sensitive search"), self
            )
        
        # The checkbox is unset initially.
        self.case_checkbox.setChecked(0)
        
        # Create labels to accompany the colour buttons.
        fg_label = QLabel(self)
        fg_label.setText(
            i18n("Foreground colour")
            )
        bg_label = QLabel(self)
        bg_label.setText(
            i18n("Background colour")
            )
        
        # Add buttons to use to specify the colour of highlighted text and
        # its background.
        self.fg_button = KColorButton(Qt.black, self)
        self.bg_button = KColorButton(Qt.yellow, self)
        
        # Add OK and Cancel buttons.
        hbox = QHBox(self)
        hbox.setSpacing(spacing)
        
        self.ok_button = KPushButton(i18n("Find"), hbox)
        self.cancel_button = KPushButton(i18n("Close"), hbox)
        
        self.connect(
            self.ok_button, SIGNAL("clicked()"), PYSIGNAL("findWord")
            )
        
        self.connect(
            self.cancel_button, SIGNAL("clicked()"), self.close
            )
        
        self.connect(
            self.line_edit, SIGNAL("returnPressed()"),
            self.ok_button.animateClick
            )
        
        # Add the widgets to the layout.
        layout.addWidget(label)
        layout.addWidget(self.line_edit)
        layout.addWidget(self.direction_combo)
        layout.addWidget(self.case_checkbox)
        layout.addWidget(fg_label)
        layout.addWidget(self.fg_button)
        layout.addWidget(bg_label)
        layout.addWidget(self.bg_button)
        layout.addWidget(hbox)
        
        # Use an accelerator to add the ability to cancel the window
        # by pressing the Escape key. Subclassing the KLineEdit widget
        # would appear to be preferable to this approach but the
        # keyPressEvent is protected, so isn't available to PyKDE.
        
        self.accel = KAccel(self)
        
        self.accel.insertItem( i18n("Close"), "Close", "Escape" )
        self.accel.connectItem( "Close", self.close )
    
    def text(self):
    
        return unicode(self.line_edit.text())
    
    def case(self):
    
        return self.case_checkbox.isChecked()
    
    def colours(self):
    
        return (
            unicode(self.fg_button.color().name()),
            unicode(self.bg_button.color().name())
            )
    
    def forwards(self):
    
        if self.direction_combo.currentItem() == 0:
            return True
        else:
            return False
