#!/usr/bin/env python

"""
console.py
Provides a Python console for Konqueror.
This module can be used with the PyAutomation KPart plugin for Konqueror.

(C) David Boddie 2004

This software is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as
published by the Free Software Foundation; either version 2 of
the License, or (at your option) any later version.

This software is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public
License along with this library; see the file COPYING
If not, write to the Free Software Foundation, Inc.,
59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
"""

from qt import QInputDialog, QLineEdit, QMouseEvent, QObject, \
               Qt, QTextBrowser, QVBoxLayout, QWidget, SIGNAL

from kdecore import \
    i18n, KApplication, KAccel, KAboutData, KCmdLineArgs, KConfig, \
    KKey, KShortcut, KURL

from khtml import KHTMLPart, DOM
from kparts import KParts

import imp, os, string, sys, traceback

# Import the __future__ module to enable code using future language changes
# to be compiled interactively.

import __future__

# Import the fastest StringIO function available.

try:

    from cStringIO import StringIO

except ImportError:

    from StringIO import StringIO


# Define the application details.

appname = u"Konqueror Python Console"
__version__ = u"0.10"


# A start up message is defined.

start_up = \
u"""
%s %s

python:
Type "help", "copyright", "credits" or "license" for more information.

konqueror:
Type "klicense" for license information specific to this environment.

""" % (appname, __version__)

# Define an object to contain the license.
klicense = unicode(__doc__[1:])

# Define a docstring for the user module.
user_docstring = unicode(__doc__[1:string.find(__doc__, "\n", 1)])

# Define a simple help function for interactive use.

help_fn = """
def help(word):

    try:
    
        return word.__doc__
    
    except AttributeError:
    
        return "No documentation found."
"""


class NoValueClass:

    pass

NoValue = NoValueClass()


class Traceback:

    def __init__(self, text):
    
        self.text = text


class Environment:

    def __init__(self):
    
        # Import some modules to customise the environment.
        self.setup_environment()
        
        # Delayed code for execution.
        self.delayed = []
    
    def setup_environment(self):
    
        # Construct a module to use to contain Python executed by the
        # user.
        self.module = imp.new_module("__user__")
        
        # Add a docstring for this module.
        self.module.__doc__ = user_docstring
        
        # Copy the sys module into the user's environment.
        self.module.sys = sys
        
        # Define some functions for the user.
        
        # Reimplement a help function like that used in Python 2.2.
        
        #help = compile(help_fn, "<setup_functions>", "exec")
        #exec help in self.module.__dict__
        
        # Set up an object containing qpython license information in the
        # user module.
        self.module.klicense = klicense
        #self.module.LICENSE = __doc__[1:]
    
    def execute(self, command_line):
    
        if len(command_line) == 0:
        
            return NoValue
        
        if command_line[0] in " \t":
        
            # Add the line to the list of delayed command lines.
            self.delayed.append(command_line)
            return NoValue
        
        if self.delayed != []:
        
            command_lines = "".join(self.delayed) + command_line
        
        else:
        
            command_lines = command_line
        
        try:
        
            value = eval(command_lines, self.module.__dict__)
            
            # Set the _ variable to the value returned.
            self.module._ = value
            
            # The pending input was executed successfully so clear
            # the delayed command lines list.
            self.delayed = []
            
            return value
        
        except SyntaxError:
        
            # Catch Syntax Errors and pass them on for another attempt
            # to compile and execute them.
            pass
        
        except:
        
            # Catch all other errors and report them.
            
            # Clear the delayed command line list.
            self.delayed = []
            return self.format_exc()
        
        
        # Try again with the exec command.
        
        # Construct a flags word for any features imported from the
        # __future__ module. Peculiarly, this has to be done because
        # the compile function appears not to be aware of any __future__
        # imports done from the command line.
        
        flags = 0
        user_values = filter(
            lambda x: isinstance(x, __future__._Feature),
            self.module.__dict__.values()
            )
        
        for value in __future__.__dict__.values():
        
            if isinstance(value, __future__._Feature):
            
                if value in user_values:
                
                    flags = flags | value.compiler_flag
        
        try:
        
            # Compile the code using the flags determined from the imported
            # features. The last parameter indicates that these flags should
            # be taken into account when compiling.
            
            code = compile(command_lines, "__user__", "exec", flags, 0)
            exec code in self.module.__dict__
            
            # The pending input was executed successfully so clear
            # the delayed command lines list.
            self.delayed = []
            
            # No values are returned from the execution of code in this
            # manner unless we are using existing lines from the delayed
            # lines list.
            return NoValue
        
        except SyntaxError:
        
            msg = sys.exc_info()[1].msg
            #print encode(msg)
            
            if string.find( msg, "unexpected EOF while parsing" ) != -1:
            
                # Delay code for later execution.
                self.delayed.append(command_line)
                return NoValue
            
            elif string.find(msg, "expected an indented block") != -1:

                # Only take this seriously if there is no preceding text
                # or if these last two lines have the same indentation.
                
                if self.delayed != [] and \
                    ( self.indentation(self.delayed[-1]) != \
                      self.indentation(command_line)          ):
                
                    # Keep the line for later if there is preceding text and
                    # the indentation has changed.
                    
                    self.delayed.append(command_line)
                    return NoValue
            
            # An actual Syntax Error.
            
            # Clear the delayed command line list.
            self.delayed = []
            return self.format_exc()
        
        except:
        
            # Other errors are caught here.
            
            # Clear the delayed command line list.
            self.delayed = []
            return self.format_exc()
    
    def format_exc(self):
    
        lines = map(
            lambda line: line,
            #traceback.format_exception_only(*(sys.exc_info()[:2]))
            traceback.format_exception(*(sys.exc_info()))
            )
        
        return Traceback("\n".join(lines))
    
    def indentation(self, s, tabsize = 4):
    
        l = 0
        
        for i in s:
        
            if i == " ":
                l = l + 1
            elif i == "\t":
                l = l + tabsize
            else:
                break
        
        return l


class Buffer:

    def __init__(self, *args):
    
        self.StringIO = StringIO(*args)
        self.read_callback = self.null_read_callback
        self.write_callback = self.null_write_callback
    
    def null_read_callback(self, buf, data):
    
        return data
    
    def null_write_callback(self, buf, data, flush = 0):
    
        pass
    
    def write(self, data):
    
        # Use another method to actually write the data without triggering
        # a callback.
        self._write(data)
        
        # Call callback.
        self.write_callback(self, data)
    
    def _write(self, data):
    
        self.StringIO.write(data)
    
    def read(self, *args):
    
        # Use another method to actually read the data without triggering
        # a callback.
        data = self._read(*args)
        
        # Call callback with the data in the buffer, giving the callback
        # the chance to modify it before its use.
        data = self.read_callback(self, data)
        
        return data
    
    def _read(self, *args):
    
        return self.StringIO.read(*args)
    
    def clear(self):
    
        self.StringIO.seek(0, 0)
        self.StringIO.truncate()
    
    def flush(self):
    
        self.write_callback(self, u"", flush = 1)
        self.StringIO.flush()
#    
#    def isatty(self):
#    
#        self.StringIO.isatty()
#    
#    def getvalue(self, *args):
#    
#        self.StringIO.getvalue(*args)
#    
    def readline(self):
    
        # Use another method to actually read the data without triggering
        # a callback.
        data = self._readline()
        
        # Call callback with the data in the buffer, giving the callback
        # the chance to modify it before its use.
        data = self.read_callback(self, data)
        
        return data
    
    def _readline(self):
    
        return self.StringIO.readline()
    
    def readlines(self):
    
        # Use another method to actually read the data without triggering
        # a callback.
        lines = self._readlines()
        
        new_lines = []
        for line in lines:
        
            # Call callback.
            new_line = self.read_callback(self, data[:-1]) + u"\n"
            new_lines.append(new_line)
        
        return new_lines
    
    def _readlines(self):
    
        return self.StringIO.readlines()
    
    def seek(self, *args):
    
        self.StringIO.seek(*args)
    
    def tell(self):
    
        return self.StringIO.tell()
    
    def writelines(self, *args):
    
        # Use another method to actually write the data without triggering
        # a callback.
        self._writelines(*args)
        
        # Call callback.
        self.write_callback(self, u"\n".join(args))
    
    def _writelines(self, *args):
    
        self.StringIO.writelines(*args)
    
    def truncate(self, *args):
    
        if args != ():
        
            self.StringIO.seek(args[0])
        
        self.StringIO.truncate()


class LineEdit(QLineEdit):

    def __init__(self, parent, slot):
    
        QLineEdit.__init__(self, parent)
        
        QObject.connect(
            self, SIGNAL("returnPressed()"), self.enter_line
            )
        
        #self.useGlobalKeyBindings()
        
        self.slot = slot
        
        self.history = []
        self.history_pos = len(self.history)
    
    def history_previous(self):
    
        self.history_pos = max(0, self.history_pos - 1)
        
        if self.history_pos < len(self.history):
        
            previous = self.history[self.history_pos]
        
        else:
        
            previous = ""
        
        self.setText(previous)
    
    def history_next(self):
    
        items = len(self.history)
        
        self.history_pos = min(self.history_pos + 1, items)
        
        if self.history_pos < items:
        
            next = self.history[self.history_pos]
        
        else:
        
            next = ""
        
        self.setText(next)
    
    def enter_line(self):
    
        self.history.append(self.text())
        self.history_pos = len(self.history)
        
        # Call the slot passed by the parent object.
        self.slot()


class Console(QWidget):

    def __init__(self, part):
    
        QWidget.__init__(self)
        
        # Create a layout widget to use to contain the HTML display and
        # the prompt widget.
        self.layout = QVBoxLayout(self)
        
        self.output = QTextBrowser(self)
        
        self.prompt = LineEdit(self, self.enter_line)
        
        self.layout.addWidget(self.output)
        self.layout.addWidget(self.prompt)
        
        # Create an environment to run Python scripts in.
        self.environment = Environment()
        
        # Redirect stdin, stdout and stderr to suitable objects.
        self.old_stdin = sys.__stdin__
        stdin = Buffer()
        stdin.read_callback = self.process_stdin
        sys.__stdin__ = sys.stdin = stdin
        
        self.old_stdout = sys.__stdout__
        stdout = Buffer()
        stdout.write_callback = self.process_stdout
        sys.__stdout__ = sys.stdout = stdout
        
        self.old_stderr = sys.__stderr__
        stderr = Buffer()
        stderr.write_callback = self.process_stderr
        sys.__stderr__ = sys.stderr = stderr
        
        # Add a new page function to the environment module.
        self.environment.module.new_page = self.new_page
        
        # Add a save output function to the environment module.
        self.environment.module.save_page = self.save_page
        
        # Add the reference to the KPart to the environment.
        self.environment.module.part = part
        
        # Start writing output to the HTML document.
        self.new_page()
        
        # Print a welcome message.
        self.add_output(start_up)
        
        # Configuration defaults
        
        self.line_keys = KAccel(self.prompt)
        
        self.line_keys.insert(
            "Previous line", i18n("Previous line"),
            i18n("Retrieve the previous line in the command history buffer."),
            KShortcut("Up"),
            self.prompt.history_previous
            )
        
        self.line_keys.insert(
            "Next line", i18n("Next line"),
            i18n("Retrieve the next line in the command history buffer."),
            KShortcut("Down"),
            self.prompt.history_next
            )
        
        self.display_width = 80
        
        # Use a counter to generate anchors so that the view can scroll.
    
    def __del__(self):
    
        sys.__stdin__ = sys.stdin = self.old_stdin
        sys.__stdout__ = sys.stdout = self.old_stdout
        sys.__stderr__ = sys.stderr = self.old_stderr
        
        open("/tmp/debug.txt", "w").write("__del__\n")
        
        QWidget.__del__(self)
    
    def new_page(self):
    
        self.output.setText("")
    
    def enter_line(self):
    
        text = unicode(self.prompt.text())
        
        # Echo the text submitted to the output window.
        self.add_output(text + u"\n")
        # Examine the text submitted.
        
        # Submit it to the user's running environment.
        try:
        
            # Retrieve a string value from the environment's execute method.
            value = self.environment.execute(text+"\n")
            
            # If no exception was raised then clear the prompt text.
            self.prompt.clear()
            
            if isinstance(value, Traceback):
            
                self.add_output(value.text + "\n")
            
            elif value != NoValue and value is not None:
            
                self.add_output(repr(value)+"\n")
        
        except SyntaxError:
        
            pass
    
    def add_output(self, output):
    
        self.output.setText(unicode(self.output.text()) + output)
        
        # Examine the height of the scroll area and ensure that the output
        # is visible.
        #width = self.output.contentsWidth()
        height = self.output.contentsHeight()
        #x = self.output.contentsX()
        #y = self.output.contentsY()
        #vwidth = self.output.visibleWidth()
        #vheight = self.output.visibleHeight()
        self.output.ensureVisible(0, height)
    
    def save_page(self, filename):
    
        """save_page(filename)
        
        Save the HTML output to the file specified.
        """
        
        # Write a terminated version of the HTML.
        open(filename, "w").write(
            unicode(self.output.text())
            )
    
    def end_output(self):
    
        pass
    
    def process_stdin(self, buf, data):
    
        # Flush stdout and stderr before obtaining input.
        sys.stdout.flush()
        sys.stderr.flush()
        
        new_data, ok = QInputDialog.getText(
            appname, "Request for input", QLineEdit.Normal, data,
            self
            )
        
        if ok == 1:
        
            # Echo stdin on stdout.
            sys.stdout.write(unicode(new_data)+"\n")
            
            return unicode(new_data)+"\n"
        
        else:
        
            return unicode(data)
    
    def process_output(self, buf, data, flush):
    
        # Take the data recently added to the buffer and add it to the
        # output.
        
        if flush == 1 or data[-1:] == u"\n":
        
            buf.seek(0, 0)
            
            # Use the public read method to extract the actual data.
            data = buf.read()
            
            #if flush == 0 and (data[:-1] == "\n" and data != "\n"):
            #
            #    # Remove trailing newlines from strings which contain
            #    # other characters, leaving single newlines alone.
            #    data = data[:-1]
            
            # Reformat, taking carriage return characters into account.
            lines = string.split(data, u"\n")
            
            new_lines = []
            
            for line in lines:
            
                # Split the line at each carriage return encountered.
                words = string.split(line, u"\r")
                
                # Construct a new line by overlaying all the "words"
                # which make up the line.
                length = reduce(lambda x,y: max(x, len(y)), words, 0)
                new_line = u" " * min(self.display_width, length)
                
                for word in words:
                
                    # Simply overlay the word over the current line string.
                    # This is not optimal if the words are very long; in that
                    # case they should be wrapped at a specified column number.
                    new_line = word + new_line[len(word):]
                
                new_lines.append(new_line)
            
            data = string.join(new_lines, u"\n")
            
            return unicode(data)
        
        return None
    
    def process_stdout(self, buf, data, flush = 0):
    
        output = self.process_output(buf, data, flush)
        
        if output is not None:
        
            self.add_output(output)
            
            # Clear the buffer.
            buf.clear()
    
    def process_stderr(self, buf, data, flush = 0):
    
        output = self.process_output(buf, data, flush)
        
        if output is not None:
        
            self.add_output(output)
            
            # Clear the buffer.
            buf.clear()


# Main function

def main(part):

    # Create a Console object, supplying it with a reference to the KPart
    # currently in use.
    global console
    
    console = Console(part)
    console.show()


def shutdown():

    global console
    
    console.close(1)
    del console
