#!/usr/bin/env python

"""
highlighter.py
Highlights all instances of the word "Python" in a web page.
This module can be used with the PyAutomation KPart plugin for Konqueror.

(C) David Boddie 2004

This software is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as
published by the Free Software Foundation; either version 2 of
the License, or (at your option) any later version.

This software is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public
License along with this library; see the file COPYING
If not, write to the Free Software Foundation, Inc.,
59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
"""

from khtml import DOM

import sys, traceback

Search_Keyword = "Python"
Case_Sensitive = 0
Foreground_Colour = "black"
Background_Colour = "yellow"


# Debugging flag: set to 1 to enable debugging.

DEBUG = 1

def debug(msg):

    if DEBUG == 0: return
    
    sys.stderr.write(msg+"\n")


# Create a trace method to catch errors in methods.

def trace(fn, args):

    debug("tracing %s%s" % (fn.__name__, args))
    try:
    
        fn(*args)
    
    except:
    
        traceback.print_exception(
            *(sys.exc_info()), **{"file": sys.stderr}
            )
    
    debug("exiting %s" % fn.__name__)

# The trap function allows functions to be called under conditions where
# any exceptions raised are caught and written to the debug file.

def trap(fn):

    """fn = trap(fn)
    
    Create a function wrapper in which the original function is called
    within a try...except clause for debugging purposes.
    """

    old_fn = fn
    return lambda *args: trace(old_fn, args)

# End of debugging functions --------------------------------------------------

def main(part):

    # Find the part's document.
    dom = part.document()
    
    # Create an object to perform the work.
    highlighter = Highlighter(dom)
    
    # Process the document.
    highlighter.highlight(
        Search_Keyword, Case_Sensitive, (Foreground_Colour, Background_Colour)
        )

main = trap(main)

class Highlighter:

    def __init__(self, dom):
    
        self.dom = dom
    
    def highlight(self, text, case, colours):
    
        root = self.dom.documentElement()
        
        body = self.find_body(root)
        
        if body is None:
        
            return
        
        self.find_matching(body, text, case, colours)
    
    def find_body(self, node):
    
        name, value, attributes = self.node_info(node)
        
        if name.upper() == u"BODY":
        
            return node
        
        # Examine any child nodes.
        
        child = node.firstChild()
        
        while not child.isNull():
        
            found = self.find_body(child)
            
            if found: return found
            
            child = child.nextSibling()
        
        return None
    
    def find_matching(self, node, text, case, colours):
    
        # Examine any child nodes.
        
        child = node.firstChild()
        
        while not child.isNull():
        
            # Retrieve the child node from the matching method.
            
            # This is performed in case the child node is replaced by
            # a set of nodes for highlighted text. Otherwise, the original
            # child node may have been removed from the document and we
            # would lose track of our position.
            child = self.find_matching(child, text, case, colours)
            
            child = child.nextSibling()
        
        if node.nodeType() != DOM.Node.TEXT_NODE:
        
            # Not a text node or it contains no text.
            
            # Return the existing node to the calling function.
            return node
        
        name, value, attributes = self.node_info(node)
        
        # Look for matching text.
        begin_at = self.match(value, text, 0, case)
        
        if begin_at != -1:
        
            # Matching text was found.
            
            # Replace this node with a collection in which the text given
            # is highlighted using <span>...</span> elements.
            
            span = self.dom.createElement( DOM.DOMString("span") )
            
            if begin_at == 0 and len(value) == len(text):
            
                # The node matches exactly.
                
                highlighted_span = self.dom.createElement( DOM.DOMString("span") )
                highlighted_span.setAttribute(
                    DOM.DOMString(u"style"),
                    DOM.DOMString(
                        u"color: %s; background-color: %s" % colours
                        )
                    )
                
                text_node = self.dom.createTextNode( DOM.DOMString(value) )
                self.set_attributes(text_node, attributes)
                
                highlighted_span.appendChild(text_node)
                
                span.appendChild(highlighted_span)
            
            else:
            
                # Split the node into pieces, if necessary.
                
                # Start looking for matches from the beginning of the
                # text node's value to allow preceding non-matching text
                # to be incorporated into the new set of nodes.
                
                last_end = 0
                
                while last_end < len(value):
                
                    begin = self.match(value, text, last_end, case)
                    
                    if begin > last_end:
                    
                        # Include the text up to the next match.
                        
                        text_node = self.dom.createTextNode(
                            DOM.DOMString(value[last_end:begin])
                            )
                        
                        self.set_attributes(text_node, attributes)
                        span.appendChild(text_node)
                    
                    elif begin == -1:
                    
                        # No more instances found.
                        
                        # Include the rest of the text found and exit the
                        # loop.
                        
                        text_node = self.dom.createTextNode(
                            DOM.DOMString(value[last_end:])
                            )
                        
                        self.set_attributes(text_node, attributes)
                        span.appendChild(text_node)
                        
                        break
                    
                    # Wrap the highlighted keyword in a <span>...</span> pair.
                    
                    highlighted_span = self.dom.createElement(
                        DOM.DOMString("span")
                        )
                    highlighted_span.setAttribute(
                        DOM.DOMString("style"),
                        DOM.DOMString(
                            u"color: %s; background-color: %s" % colours
                            )
                        )
                    
                    end = begin + len(text)
                    
                    text_node = self.dom.createTextNode(
                        DOM.DOMString(value[begin:end])
                        )
                    
                    self.set_attributes(text_node, attributes)
                    highlighted_span.appendChild(text_node)
                    
                    span.appendChild(highlighted_span)
                    
                    # Record the end of the match found, ready for the next
                    # search.
                    last_end = end
            
            # Replace the current node.
            parent = node.parentNode()
            parent.replaceChild(span, node)
            
            # Return the new set of nodes, contained within the span
            # element to the calling function.
            return span
        
        # Return the existing node to the calling function.
        return node
    
    def match(self, value, text, start, case):
    
        if case:
        
            return value.find(text, start)
        
        else:
        
            return value.upper().find(text.upper(), start)
    
    def node_info(self, node):
    
        name = unicode(node.nodeName().string())
        value = unicode(node.nodeValue().string())
        
        attrs = node.attributes()
        attributes = {}
        
        for a in range(0, attrs.length()):
        
            item = attrs.item(a)
            attr_name = unicode(item.nodeName().string()).lower()
            attributes[attr_name] = unicode(item.nodeValue().string())
        
        return name, value, attributes
    
    def set_attributes(self, node, attributes):
    
        for name, value in attributes.items():
        
            node.setAttribute( DOM.DOMString(name), DOM.DOMString(value) )
