#!/usr/bin/env python

"""
page-summary.py
Displays a summary of the currently displayed web page.
This module can be used with the PyAutomation KPart plugin for Konqueror.

(C) David Boddie 2004

This software is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as
published by the Free Software Foundation; either version 2 of
the License, or (at your option) any later version.

This software is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public
License along with this library; see the file COPYING
If not, write to the Free Software Foundation, Inc.,
59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
"""

from kdecore import KURL

from khtml import DOM

# Import Paul's khtmldom interface to KHTML for a more Pythonic DOM
# experience.

import khtmldom

import sys


def main(part):

    # Retrieve the document object.
    document = part.document()
    
    # Obtain a wrapped form of the document.
    document = khtmldom.Node(document)
    
    # Find the document body.
    bodies = find_nodes(document, u"BODY", case = 0)
    
    try:
    
        body = bodies[0]
    
    except IndexError:
    
        show_error_page(part, u"Failed to find a document body.")
        return
    
    # Find the base URL for links.
    
    try:
    
        base_node = find_nodes(document, u"BASE", case = 0)[0]
        base = base_node.attributes[(u"HREF")].nodeValue
    
    except (IndexError, KeyError, AttributeError):
    
        base = str(part.url())
    
    # Find the title of the page.
    
    try:
    
        title_element = find_nodes(document, u"TITLE", case = 0)[0]
        title = u"".join(
            map(
                lambda piece: piece.nodeValue,
                find_nodes(title_element, u"#text", case = 0)
                )
            )
    
    except (IndexError, AttributeError):
    
        title = str(part.url())
    
    # Find paragraph elements.
    
    stories = []
    
    for paragraph in find_nodes(body, u"P", case = 0):
    
        # Examine each paragraph to see whether it contains an anchor.
        
        for anchor in find_nodes(paragraph, u"A", case = 0):
        
            try:
            
                # Read the name or href attributes of the element.
                name = anchor.getAttribute(u"NAME")
                href = anchor.getAttribute(u"HREF")
            
            except (KeyError, AttributeError):
            
                # Try the next anchor.
                continue
            
            else:
            
                # Read all the text found within the anchor element.
                pieces = find_nodes(anchor, u"#text", case = 0)
                
                link_text = u"\n".join(
                    map(lambda piece: piece.nodeValue, pieces)
                    )
                
                if name:
                    link = base + name
                elif href:
                    link = href
                else:
                    link = base
            
            # Read all the paragraph text as well.
            text = serialise(paragraph)
            
            # Put the anchor name and text in the list of stories.
            stories.append( (link, link_text, text) )
    
    summaries = (
        u"\n<ul>"
        u"\n<dl>" + \
        u"\n<dt>".join(
            map(
                lambda story:
                    u'<a href="%s">%s</a></dt>\n<dd>%s</dd>\n' % story,
                stories
                )
            ) + \
        u"\n</p>"
        )
    
    # Write a new page containing the summaries.
    part.begin()
    
    part.write(
        u"\n".join(
            [
                u"<html>",
                u"<head>",
                u"  <title>Summary of %s</title>" % title,
                u'  <meta http-equiv="Content-Type" content="text/html">\n',
                u"</head>",
                u"",
                u"<body>",
                u"",
                u"<h1>Summary of %s</h1>" % title,
                u"",
                summaries,
                u"",
                u"</body>",
                u"</html>",
                u""
            ]
            )
        )
    
    part.end()


def show_error_page(part, message):

    # Clear the document and replace it with some information.
    
    part.begin()
    
    part.write( u"\n".join(
        [
            u"<html>",
            u"<head>",
            u"  <title>Message from script</title>",
            u'  <meta http-equiv="Content-Type" content="text/html">\n',
            u"</head>",
            u"",
            u"<body>",
            u"<h1>Message from script</h1>",
            u"",
            message,
            u"",
            u"</body>",
            u"</html>",
            u""
        ]
        ) )
    
    part.end()


def find_nodes(node, name, case = 1):

    # Find all nodes with the given name under the named node in the
    # document, respecting case sensitivity if case is set to one.
    
    found = []
    
    for child in node.childNodes:
    
        # Check this node to determine whether it matches the name
        # given.
        
        if case == 1 and child.nodeName == name:
        
            found.append(child)
        
        elif case == 0 and child.nodeName.lower() == name.lower():
        
            found.append(child)
        
        # Check this node's children, too.
        
        found = found + find_nodes(child, name, case)
    
    return found

def serialise(node):

    found = []
    
    for child in node.childNodes:
    
        # Check this node to determine whether it matches the name
        # given.
        
        if str(child.nodeName) == u"#text":
        
            output = [child.nodeValue]
        
        else:
        
            output = [u"<"+child.nodeName+u" "]
            for ns, attr in child.attributes.keys():
            
                output.append(attr+u"="+child.getAttribute(attr)+u" ")
            
            output.append(u">")
            
            if child.childNodes != []:
            
                # Check this node's children, too.
                
                output.append(serialise(child))
                
                output.append(u"</"+child.nodeName+u">")
        
        found.append( u"".join(output) )
    
    return u"".join(found)

def html_repr(text):

    new = []
    
    for c in text:
    
        if c == u"&":
            new.append(u"&amp;")
        elif c == u"<":
            new.append(u"&lt;")
        elif c == u">":
            new.append(u"&gt;")
        else:
            new.append(c)
    
    return u"".join(new)
