#!/usr/bin/env python

"""
reformat.py
Reformats preformatted text on web pages to a predetermined line width.
This module can be used with the PyAutomation KPart plugin for Konqueror.

(C) David Boddie 2005

This software is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as
published by the Free Software Foundation; either version 2 of
the License, or (at your option) any later version.

This software is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public
License along with this library; see the file COPYING
If not, write to the Free Software Foundation, Inc.,
59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
"""

from qt import QString
from khtml import DOM

import sys, traceback

LineWidth = 77


# Debugging flag: set to 1 to enable debugging.

DEBUG = 1
DEBUG_FILE = "/tmp/pydebug.txt"

def debug(msg):

    if DEBUG == 0: return
    open(DEBUG_FILE, "a").write(msg+"\n")
    sys.stderr.write(msg+"\n")


# Create a trace method to catch errors in methods.

def trace(fn, args):

    debug("tracing %s%s" % (fn.__name__, args))
    try:
    
        result = fn(*args)
        debug("result: %s" % repr(result))
        return result
    
    except:
    
        traceback.print_exception(
            *(sys.exc_info()), **{"file": open(DEBUG_FILE, "a")}
            )
        traceback.print_exception(
            *(sys.exc_info()), **{"file": sys.stderr}
            )
    
    debug("exiting %s" % fn.__name__)

# The trap function allows functions to be called under conditions where
# any exceptions raised are caught and written to the debug file.

def trap(fn):

    """fn = trap(fn)
    
    Create a function wrapper in which the original function is called
    within a try...except clause for debugging purposes.
    """

    old_fn = fn
    return lambda *args: trace(old_fn, args)

# End of debugging functions --------------------------------------------------

def main(part):

    # Find the part's document.
    dom = part.document()
    
    # Create an object to perform the work.
    formatter = Formatter(dom)
    
    # Process the document.
    formatter.reformat(LineWidth)

main = trap(main)

class Formatter:

    def __init__(self, dom):
    
        self.dom = dom
    
    def reformat(self, line_width):
    
        root = self.dom.documentElement()
        
        body = self.find_body(root)
        
        if body is None:
        
            return
        
        self.find_preformatted(body, line_width)
    
    def find_body(self, node):
    
        name, value, attributes = self.node_info(node)
        
        if name.upper() == u"BODY":
        
            return node
        
        # Examine any child nodes.
        
        child = node.firstChild()
        
        while not child.isNull():
        
            found = self.find_body(child)
            
            if found: return found
            
            child = child.nextSibling()
        
        return None
    
    def find_preformatted(self, node, line_width):
    
        # Examine any child nodes.
        
        if node.nodeName().string() == QString("PRE"):
        
            # A preformatted text node.
            self.split_text(node, line_width, 0)
            return
        
        child = node.firstChild()
        
        while not child.isNull():
        
            # Retrieve the child node from the matching method.
            
            # This is performed in case the child node is replaced by
            # a set of nodes for highlighted text. Otherwise, the original
            # child node may have been removed from the document and we
            # would lose track of our position.
            self.find_preformatted(child, line_width)
            
            child = child.nextSibling()
    
    def split_text(self, parent, line_width, cursor):
    
        child = parent.firstChild()
        
        while not child.isNull():
        
            if child.nodeType() == DOM.Node.TEXT_NODE:
            
                text = unicode(child.nodeValue().string())
                lines = text.split(u"\n")
                lengths = []
                length = cursor
                
                for line in lines:
                
                    length = length + len(line)
                    lengths.append(length)
                    length = 0
                
                if max(lengths) > line_width:
                
                    # Replace the current node.
                    span = self.dom.createElement(DOM.DOMString("span"))
                    
                    for i in range(len(lines)):
                    
                        line = lines[i]
                        indent = len(line)-len(line.lstrip())
                        if indent > 0:
                        
                            words = line.lstrip().split()
                            if words != []:
                                words[0] = u" "*indent + words[0]
                        
                        else:
                        
                            words = line.split()
                        
                        length = lengths[i]
                        
                        while words != []:
                        
                            if cursor + len(words[0]) <= line_width:
                            
                                word = words.pop(0)
                                cursor = cursor + len(word)
                                
                                span.appendChild(
                                    self.dom.createTextNode(DOM.DOMString(word))
                                    )
                                
                                if words != [] and cursor + 1 < line_width:
                                
                                    span.appendChild(
                                        self.dom.createTextNode(
                                            DOM.DOMString(u" ")
                                            )
                                        )
                                    cursor = cursor + 1
                            
                            elif cursor == 0:
                            
                                word = words[0][:line_width]
                                words[0] = words[0][line_width:]
                                cursor = 0
                                
                                span.appendChild(
                                    self.dom.createTextNode(DOM.DOMString(word))
                                    )
                            
                            else:
                            
                                span.appendChild(
                                    self.dom.createTextNode(DOM.DOMString(u"\n"))
                                    )
                                cursor = 0
                        
                        span.appendChild(
                            self.dom.createTextNode(DOM.DOMString(u"\n"))
                            )
                        cursor = 0
                    
                    next = child.nextSibling()
                    parent.replaceChild(span, child)
                    debug("ending with cursor = %i" % cursor)
                
                else:
                
                    cursor = lengths[-1]
                    next = child.nextSibling()
                
                child = next
            
            elif child.hasChildNodes():
            
                cursor = self.split_text(child, line_width, cursor)
                child = child.nextSibling()
            
            else:
            
                child = child.nextSibling()
        
        return cursor
    
    #split_text = trap(split_text)
    
    def node_info(self, node):
    
        name = unicode(node.nodeName().string())
        value = unicode(node.nodeValue().string())
        
        attrs = node.attributes()
        attributes = {}
        
        for a in range(0, attrs.length()):
        
            item = attrs.item(a)
            attr_name = unicode(item.nodeName().string()).lower()
            attributes[attr_name] = unicode(item.nodeValue().string())
        
        return name, value, attributes
    
    def set_attributes(self, node, attributes):
    
        for name, value in attributes.items():
        
            node.setAttribute( DOM.DOMString(name), DOM.DOMString(value) )
