/*
 * pykpart_launcher.cpp
 * 
 * Launch KParts and KPart plugins written in Python using an embedded Python
 * interpreter.
 * 
 * Copyright (C) 2003-2005 Jim Bublitz, David Boddie <david@boddie.org.uk>
 * 
 * This software is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public
 * License along with this library; see the file COPYING
 * If not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include <Python.h>
#include <qobject.h>
#include <klibloader.h>
#include "pykpart_launcher.h"

#define debug 0

void* report_error(char *msg, bool report_python = true)
{
    #if debug == 1
    printf ("error: %s\n", msg);
    #endif

    if (report_python) PyErr_Print();

    return NULL;
}

PyObject* call_function(PyObject *callable, PyObject *args)
{
    PyObject *result = NULL, *pArgs;

    if (callable == NULL)
    {
        return NULL;
    }

    if (PyCallable_Check(callable))
    {
        if (args == NULL)
            pArgs = PyTuple_New(0);
        else
            pArgs = args;

        result = PyObject_CallObject(callable, pArgs);

        /* If the arguments were created in this function then decrease
           their reference count. */
        if (args == NULL)
            Py_XDECREF(pArgs);
            /* pDict and pFunc are borrowed and must not be Py_DECREF-ed */

        if (result == NULL)
        {
            #if debug == 1
            PyErr_Print();
            #else
            PyErr_Clear();
            #endif
        }
    }

    return result;
}

#if debug == 1

static PyObject * kpartdebug_showParents(PyObject *self, PyObject *args)
{
    if (PyTuple_Size(args) != 1)
        return NULL;

    PyObject *pyobject = PyTuple_GetItem(args, 0);
    QObject *qobject = (QObject*)PyLong_AsVoidPtr(pyobject);
    fprintf(stderr, "parent: %s\n", qobject->className());

    Py_INCREF(Py_None);
    return Py_None;
}

static PyMethodDef kpartdebug_methods[] = {
    { "showParents", kpartdebug_showParents, METH_VARARGS,
      "showParents(object)\n\nShow the parents of the given object."},
    { NULL, NULL } /* sentinel */
};

void init_kpartdebug(void)
{
    PyImport_AddModule("kpartdebug");
    Py_InitModule("kpartdebug", kpartdebug_methods);
}

#endif

// Declare a Python object to represent the object returned by the
// Python module.
static KLibFactory *pluginFactory = NULL;

extern "C"
{
    #if debug == 1
    void initkpartdebug(void)
    {
        init_kpartdebug();
    }
    #endif

    void* CPP_FACTORY(void)
    {
        if (!Py_IsInitialized())
        {
            KLibLoader::self()->globalLibrary(PYTHON_LIB_PATH);
            Py_Initialize();
        }

        #if debug == 1
        init_kpartdebug();
        #endif
        
        // Locate the path to the script and append it to sys.path list.

        PyRun_SimpleString("import sys\n");
        PyRun_SimpleString("sys.path.append('"MODULE_DIR"')\n");

        PyObject *pyModule = PyImport_ImportModule(MODULE_NAME);
        /* pyModule is a new reference */

        if (!pyModule) return report_error ("***failed to import module\n");

        PyObject *pyDict = PyModule_GetDict(pyModule);
        /* pDict is a borrowed reference */

        Py_XDECREF(pyModule);

        PyRun_String("import sip\n"
                     "def __embedded_factory__init__(parent = None, name = None):\n"
                     "    global factory\n"
                     "    factory = "FACTORY"(parent, name)\n"
                     "    return sip.unwrapinstance(factory)\n",
                     Py_file_input, pyDict, pyDict);

        // Get the Python module's factory function from the header file at
        // compile time.
        PyObject *pyFactory = PyDict_GetItemString(pyDict, "__embedded_factory__init__");
        /* pyFactory is a borrowed reference */

        // Set up the args for the Python factory function call.
        PyObject *args = NULL;

        // Instantiate the class.
        PyObject *pyFactoryPtr = call_function (pyFactory, args);
        /* pyFactoryPtr should be a new reference to a long object. */

        if (pyFactoryPtr)
        {
            // Convert the PyObject to a C++ object. *
            pluginFactory = (KLibFactory*)PyLong_AsVoidPtr(pyFactoryPtr);
        }

        Py_XDECREF(pyFactoryPtr);

        return pluginFactory;
    }
}
