#!/usr/bin/env python

"""
scrapbook.py

KPart plugin for KHTML to allow text in an HTML page to be extracted and
placed in a separate window.

Copyright (C) 2005 David Boddie <david@boddie.org.uk>

This software is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as
published by the Free Software Foundation; either version 2 of
the License, or (at your option) any later version.

This software is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public
License along with this library; see the file COPYING
If not, write to the Free Software Foundation, Inc.,
59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

"""

# Import the required Qt and KDE classes.

from qt import QCursor, QMessageBox, QPopupMenu, QString, QStringList, Qt, \
               QWidget, PYSIGNAL, SIGNAL

from kdecore import i18n, KAccel, KInstance, KLibFactory, KShortcut, KURL
from kparts import KParts
from kdeui import KAction, KPushButton, KTabBar, KTabWidget
from khtml import KHTMLPart, DOM


# Define a class which will be initialised when the plugin is loaded.

class FactoryClass(KLibFactory):

    instance = None
    
    def __init__(self, parent = None, name = None):
    
        KLibFactory.__init__(self, parent, name)
        
        if FactoryClass.instance is None:
        
            FactoryClass.instance = KInstance("Scrapbook")
    
    def createObject(self, parent = None, name = None, class_name = "QObject",
                     args = None):
    
        if args is None:
        
            args = []
        
        plugin = Scrapbook(parent, name)
        self.emit(SIGNAL("objectCreated(QObject)"), (plugin,))
        return plugin


# Define the plugin class, derived from the standard KParts.Plugin class.

class Scrapbook(KParts.Plugin):

    def __init__(self, parent = None, name = None):
    
        KParts.Plugin.__init__(self, parent, name)
        
        self._action = KAction(
            i18n("Copy to scrapbook"),
            "scrapbook",
            KShortcut(i18n("Ctrl+Shift+C")),
            self.copyToScrapbook,
            self.actionCollection(),
            "extracthtml"
            )
        self.scrapbook = ScrapbookWidget()
    
    def copyToScrapbook(self):
    
        # The parent is assumed to be a KHTMLPart
        
        khtmlpart = self.parent()

        if not khtmlpart.inherits("KHTMLPart"):
        
            title = i18n("Cannot copy text")
            text = i18n("Only HTML can be copied to the scrapbook.")
            
            QMessageBox.warning( None, title, text )
            return
        
        selection = khtmlpart.selection()
        if selection.isNull():
        
            text = khtmlpart.document().toString().string()
        
        else:
        
            html = selection.toHTML()
            text = html.string()
        
        url = khtmlpart.url()
        
        titleNodes = khtmlpart.document().getElementsByTagName(DOM.DOMString("title"))
        title = url.prettyURL()
        
        if titleNodes.length() >= 1:
            children = titleNodes.item(0).childNodes()
            if children.length() >= 1:
                title = children.item(0).nodeValue().string()
        
        title = title.simplifyWhiteSpace()
        self.scrapbook.addCutting(title, url, text)
        
        if not self.scrapbook.isShown():
            self.scrapbook.show()


class ScrapbookWidget(KTabWidget):

    def __init__(self, parent = None):
    
        self.pages = []
        KTabWidget.__init__(self, parent)
        self.setTabBar(ScrapbookTabBar(self))
        self.keys = KAccel(self)
        self.keys.insert(
            "closetab", i18n("&Close page"),
            i18n("Closes the currently opened page in the scrapbook"),
            KShortcut(i18n("Ctrl+W")),
            self.removeCutting)
        self.connect(self.tabBar(), PYSIGNAL("tabRemoved"), self.removeCutting)
        self.setCaption(i18n("Scrapbook"))
    
    def __del__(self):
    
        # Explicitly delete the pages that we manage for the tab widget.
        del self.pages
    
    def addCutting(self, title, url, text):
    
        page = KHTMLPart()
        page.setJScriptEnabled(False)
        page.setJavaEnabled(False)
        page.setStatusMessagesEnabled(False)
        page.setPluginsEnabled(False)
        page.begin(url)
        page.write(text)
        page.end()
        self.addTab(page.view(), title)
        self.pages.append((page, title, url))
        self.showPage(page.view())
    
    def removeCutting(self, index = None):
    
        if index is None:
            index = self.currentPageIndex()
        
        self.removePage(self.pages[index][0].view())
        self.pages.pop(index)


class ScrapbookTabBar(KTabBar):

    def __init__(self, parent):
    
        KTabBar.__init__(self, parent)
    
    def contextMenuEvent(self, event):
    
        position = QCursor.pos()
        tabPosition = self.mapFromGlobal(position)
        tab = self.selectTab(tabPosition)
        
        if not tab:
            return
        
        menu = QPopupMenu(self)
        removeId = menu.insertItem(i18n("&Remove Cutting"))
        
        if menu.exec_loop(position) == removeId:
        
            index = self.indexOf(tab.identifier())
            self.emit(PYSIGNAL("tabRemoved"), (index,))
