#!/usr/bin/env python

"""
setup.py

Build/install/uninstall PyKDE components which use dynamically loaded
libraries.

Copyright (C) 2003-2005 David Boddie <david@boddie.org.uk>

This software is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as
published by the Free Software Foundation; either version 2 of
the License, or (at your option) any later version.

This software is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public
License along with this library; see the file COPYING
If not, write to the Free Software Foundation, Inc.,
59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

"""

import distutils.sysconfig
import glob, imp, os, popen2, pprint, string, sys
import pykdeconfig

__version__ = "0.12 (Monday 5th September 2005)"

try:

    extsep = os.extsep

except AttributeError:

    extsep = "."


class BuildError(Exception):

    pass


# Reporting functions ---------------------------------------------------------

def inform(s):

    sys.stdout.write(s)
    sys.stdout.write("\n")
    sys.stdout.flush()

def warning(s):

    sys.stderr.write(s)
    sys.stderr.write("\n")
    sys.stderr.flush()

# File and directory functions ------------------------------------------------

def find_files(patterns, paths, return_first = 0):

    # Search each path recursively for suitable files.
    found = []
    
    for path in paths:
    
        # Find files in this path.
        found_here = []
        
        for pattern in patterns:
        
            files = glob.glob(os.path.join(path, pattern))
            
            found_here = found_here + files
        
        # Record all the files found on this path.
        if found_here != []:
        
            found.append( (path, found_here) )
            
            if return_first == 1:
            
                return found
        
        # Examine any directories on this path.
        catalogue = glob.glob(os.path.join(path, "*"))
        
        dirs = []
        
        for object in catalogue:
        
            if os.path.isdir(object) and not os.path.islink(object):
            
                # Examine each subdirectory.
                dirs.append(object)
        
        # Examine all subdirectories.
        found = found + find_files(patterns, dirs, return_first)
        
        if found != [] and return_first == 1:
        
            return found
    
    # Return a list of tuples containing the paths where suitable files are
    # found and the files in each path.
    return found

def show_directories(dirs):

    for d in dirs:
    
        inform( "    %s" % d )
    
    inform("")

def create_paths(path, files):

    paths = []
    
    for file in files:
    
        split_path = file.split("/")
        platform_path = os.sep.join(split_path)
        
        split_file = platform_path.split(".")
        platform_file = os.extsep.join(split_file)
        
        paths.append(os.path.join(path, platform_file))
    
    return tuple(paths)

def create_path(path, file):

    return create_paths(path, [file])[0]

def _path(file):

    return create_paths(build_path, [file])[0]

def install_res(source, destination, filetype = None):

    """install_res(source, destination, filetype = None)
    
    Install the resource in the named source file at the destination,
    giving the destination file a filetype if specified.
    
    If the installation fails, the BuildError exception is raised.
    """
    
    #print "install:", source, destination
    #return
    
    if not os.path.exists(source):
    
        warning(
            "Could not install %s in its target location: %s" % \
                (source, destination)
            )
        warning( "The source file does not exist." )
        return
    
    try:
    
        if filetype is None:
        
            install_type = "INSTALL"
        
        else:
        
            install_type = "INSTALL_" + string.upper(filetype)
        
        command_dict = \
            {
                "command": distutils.sysconfig.get_config_var(install_type),
                "source": source,
                "destination": destination
            }
        
        _out, _in, _err = popen2.popen3(
            "%(command)s %(source)s %(destination)s" % command_dict
            )
        
        if _err.read() != "":
        
            raise IOError
        
        #open(destination, "wb").write(
        #    open(source, "rb").read()
        #    )
        inform( "Copied %s to %s" % (source, destination) )
    
    except IOError:
    
        warning(
            "Could not install %s in its target location: %s" % \
                (source, destination)
            )
        
        raise BuildError

def uninstall_res(name, critical = 1):

    """uninstall_res(name, critical = 1)
    
    Uninstall the resource at the file with the name given.
    
    If the resource does not exist then the critical flag indicates whether
    its absence should be reported as a warning (if non-zero) or a message
    (if zero).
    
    If the uninstallation fails, the BuildError exception is raised.
    """
    
    #print "uninstall:", name
    #return
    
    if not os.path.exists(name):
    
        # Critical problems are reported as warnings. These are reported
        # on stderr and may be used by graphical installers to indicate
        # failure of the uninstallation process.
        
        if critical == 1:
            warning( "%s does not exist." % name )
        else:
            inform( "%s does not exist." % name )
        
        return
    
    try:
    
        if os.path.isdir(name):
        
            os.rmdir(name)
            inform( "Removed %s" % name )
        
        else:
        
            os.remove(name)
            inform( "Deleted %s" % name )
    
    except OSError:
    
        warning( "Failed to uninstall %s" % name )
        raise BuildError, "Failed to uninstall %s" % name

def mkdir(directory):

    """mkdir(directory)
    
    Create a directory at the path specified.
    
    If the operation fails, the BuildError exception is raised.
    """
    #print "mkdir:", directory
    #return
    
    try:
    
        os.mkdir(directory)
    
    except OSError:
    
        warning( "Could not create the support directory: %s" % path )
        raise BuildError

def symlink(src, dest):

    try:
    
        os.symlink(src, dest)
        inform( "Made symbolic link from %s to %s" % (src, dest) )
    
    except OSError:
    
        warning( "Could not make symbolic link from %s to %s" % (src, dest) )


def find_filenames_and_destinations(requires, provides, components,
                                    install = False):

    """find_filenames_and_destinations(requires, provides, components,
                                       install = False)
    
    Determine the filenames for the library, linking script and other
    resources.
    
    If the installation fails, the BuildError exception is raised.
    """
    
    filenames = \
    {
        "source":  string.join(
            string.split(requires["library"], "."), os.extsep
            ),
        "header":  string.join(
            string.split(requires["library"], ".")[:-1] + ["h"], os.extsep
            ),
        "object":  string.join(
            string.split(requires["library"], ".")[:-1] + ["o"], os.extsep
            ),
        "library": provides[".so"]["filename"] + \
            distutils.sysconfig.get_config_var("SO"),
        
        "linking": provides[".la"]["filename"] + os.extsep + "la",
    }
    
    # Determine whether each resource can be placed in the requested
    # locations.
    
    destinations = {}
    
    for suffix, item in provides.items():
    
        destination = string.join(
            string.split(item["destination"] % components, "/"), os.sep
            )
        
        if install and not os.path.isdir(destination):
        
            warning(
                'Cannot use %s as destination for "%s"' % \
                    (destination, suffix)
                )
            
            raise BuildError
        
        # Add the destination to the dictionary.
        destinations[suffix] = destination
        
        if item.has_key("filename"):
        
            filenames[suffix] = item["filename"] + os.extsep + suffix[1:]
    
    # Synonyms
    
    destinations["library"] = destinations[".so"]
    destinations["linking"] = destinations[".la"]
    
    return filenames, destinations


def make_template_dict(mod_details, requires, provides, components,
                       destinations):

    # Find the compiler and linker flags.
    compiler = distutils.sysconfig.get_config_var("CXX")
    cc_flags = "-c -g " + distutils.sysconfig.get_config_var("CFLAGS") + \
        " " + distutils.sysconfig.get_config_var("CFLAGSFORSHARED")
    
    includes = \
    [
        "-I" + distutils.sysconfig.get_config_var("INCLUDEDIR"),
        "-I" + distutils.sysconfig.get_config_var("INCLUDEPY"),
        "-I %(python inc dir)s",
        "-I %(kde inc dir)s",
        "-I %(qt inc dir)s",
        "-I."
    ]
    
    includes = string.join(includes, " ") % components
    
    
    linker = distutils.sysconfig.get_config_var("LDSHARED")
    link_flags = distutils.sysconfig.get_config_var("LINKFORSHARED") + \
                " -module -Xlinker -no-undefined " # 
    
    link_libs = \
    [
        "-L%(python dir)s -lpython%(python version)s",
        "-L%(kde lib dir)s -lkdecore",
        "-L%(qt lib dir)s -lqt-mt",
        "%(c++ libs)s",
        "%(python libs)s",
        "-Xlinker -R%(python dir)s",
        "-Xlinker -R%(qt lib dir)s",
        "-Xlinker -R%(kde lib dir)s"
    ]
    
    # Add any other required Python module libraries to the list.
    for lib, path in requires["python"]:
    
        link_libs.append( create_path(*components[lib]) )
        link_libs.append("-Xlinker -R%s" % components[lib][0])
    
    link_libs = string.join(link_libs, " ") % components
    
    # Create a dictionary containing the content required for all the
    # generated content.
    
    template_dict = \
    {
        "name":             mod_details.get("name", ""),
        "icon":             string.split(mod_details["icon"], ".")[0],
        "lib":              provides[".so"]["filename"],
        "libdest":          destinations[".so"],
        "libname":          provides[".so"]["name"],
        "compiler":         compiler,
        "cc flags":         cc_flags,
        "includes":         includes,
        "linker":           linker,
        "deps":             link_libs,
        "link flags":       link_flags,
        
        "pretty name": mod_details.get("pretty name", ""),
        "comment":  mod_details.get("comment", ""),
        "parent":   mod_details.get("parent", ""),
        "protocol": provides.get(".protocol", {"protocol": ""})["protocol"],
        "module type":  mod_details.get("module type", ""),
        "root only":  mod_details.get("root only", "false"),
        "C++ factory":  mod_details.get("C++ factory", ["", ""])[1],
        "unique":       mod_details.get("unique", "false")
    }
    
    return template_dict


# Actions ---------------------------------------------------------------------

def build(mod_details, requires, provides, components, filenames, destinations):

    # Generate content for various resources based on the templates provided.
    
    template_dict = make_template_dict(
        mod_details, requires, provides, components, destinations
        )
    
    # Generate header file for the library.

    try:

        if provides[".py"].has_key("directory"):

            path = os.path.join(
                destinations[".py"], provides[".py"]["directory"]
                )

        else:

            path = destinations[".py"]

        header_file = _path(filenames["header"])
        header = \
            '#define MODULE_DIR "%s"\n' \
            '#define MODULE_NAME "%s"\n' \
            '#define FACTORY "%s"\n' \
            '#define CPP_FACTORY %s\n' \
            '#define PYTHON_LIB_PATH "%s"\n' % (
                path, provides[".py"]["filename"],
                mod_details["factory"],
                string.join(mod_details.get("C++ factory", ["", ""]), ""),
                create_path(
                    components["python dir"], components["python shared lib"]
                    )
                )

        open(header_file, "w").write(header)

    except IOError:

        warning(
            "Could not write the header file: %s" % filenames["header"]
            )

        return

    # Compile the library.

    command = "%s %s %s %s -o %s" % (
        template_dict["compiler"], template_dict["cc flags"],
        _path(filenames["source"]),
        template_dict["includes"], _path(filenames["object"])
        )
    inform( command )

    cmd = popen2.Popen3(command, 1)

    if cmd.wait() != 0:

        warning( "Failed to compile library:" )
        warning("")
        warning( cmd.childerr.read() )
        return

    # Link the resulting object file to create a shared library.

    command = "%s %s %s %s -o %s" % (
        template_dict["linker"], template_dict["link flags"],
        _path(filenames["object"]), template_dict["deps"],
        _path(filenames["library"])
        )
    inform( command )

    cmd = popen2.Popen3(command, 1)

    if cmd.wait() != 0:

        warning( "Failed to compile library:" )
        warning("")
        warning( cmd.childerr.read() )
        return

    
    # The following procedures need to be completed even if no library
    # making was performed.
    
    for suffix, item in provides.items():
    
        if not item.has_key("contents"): continue
        
        # Generate a linking script from a template.
        
        script = item["contents"] % template_dict
        
        try:
        
            # Write the output file.
            open(_path(filenames[suffix]), "w").write(script)
        
        except IOError:
        
            warning( "Could not write the file: %s" % filenames[suffix] )
            return

def clean(mod_details, requires, provides, components, filenames, destinations):

    # Remove the intermediate files created by the build process.
    
    uninstall_res(_path(filenames["header"]))
    uninstall_res(_path(filenames["object"]))
    uninstall_res(_path(filenames["library"]))
    
    uninstall_res(_path("build.pyc"), critical = 0)
    uninstall_res(_path("details.pyc"), critical = 0)
    
    for suffix, item in provides.items():
    
        if not item.has_key("contents"): continue
        
        uninstall_res(_path(filenames[suffix]))


def install(mod_details, requires, provides, components, filenames, destinations):

    # Copy the library and linking script to the correct locations.

    install_res(
        _path(filenames["library"]),
        os.path.join(destinations["library"], filenames["library"]),
        filetype = "shared"
        )
    
    # Create the directory for the Python module, if required.
    
    if provides[".py"].has_key("directory"):
    
        path = os.path.join(
            destinations[".py"], provides[".py"]["directory"]
            )
    
    else:
    
        path = destinations[".py"]
    
    if not os.path.exists(path):
    
        mkdir(path)
    
    # Copy the Python module to the support directory.
    
    install_res(_path(filenames[".py"]), os.path.join(path, filenames[".py"]))
    
    # Install any other items in the directory along with the
    # Python module.

    # Use create_path as it is able to deal with platform specific
    # suffix separators.

    for item in requires["other items"]:

        dirpath = path

        for subdir in item.split("/")[:-1]:

            # The item is within a subdirectory which we must recreate
            # in the target location.
            dirpath = create_path(dirpath, subdir)

            # Check whether the directory exists but allow error
            # reporting for overwriting of files.

            if not os.path.isdir(dirpath):

                os.mkdir(dirpath)

        install_res(_path(item), create_path(path, item))


    # Copy the generated content to the appropriate destinations.

    for suffix, item in provides.items():

        if not item.has_key("contents"): continue

        # Retrieve the destination of the resource.
        destination = destinations[suffix]

        if item.has_key("directory"):

            directory = string.join(
                string.split(item["directory"] % components, "/"), os.sep
                )

            destination = os.path.join(destination, directory)

            if not os.path.exists(destination):

                mkdir(destination)

        # Install the resource at the destination specified.
        install_res(
            _path(filenames[suffix]),
            os.path.join(destination, filenames[suffix])
            )

    # Copy any images to the relevant directories.

    if provides.has_key("icons"):

        kde_icons_dir = destinations["icons"]

        # Install the icon for each theme in the appropriate
        # subdirectories of the icons directory.

        for theme, sizes in provides["icons"]["themes"].items():

            # Construct the path to the target subdirectory associated
            # with the theme given and check whether it exists.
            theme_dir = os.path.join(kde_icons_dir, theme)

            if os.path.isdir(theme_dir):

                # Examine each size of icon provided.

                for size, images in sizes.items():

                    # Construct the path to the target subdirectory
                    # associated with the icon size given and check
                    # whether it exists.

                    size_dir = os.path.join(
                        theme_dir, size, provides["icons"]["icon type"])

                    if os.path.isdir(size_dir):

                        # Install the image using the name given in
                        # the details module.

                        for src_name, dest_name in images:

                            install_res(
                                _path(src_name),
                                create_path(size_dir, dest_name)
                                )


def uninstall(mod_details, requires, provides, components, filenames,
              destinations):

    # Uninstall the Python module and any containing directory.

    if provides[".py"].has_key("directory"):
    
        path = os.path.join(
            destinations[".py"], provides[".py"]["directory"]
            )
    
    else:
    
        path = destinations[".py"]
    
    # Uninstall the Python module and any bytecode.
    
    uninstall_res(os.path.join(path, filenames[".py"]))
    uninstall_res(os.path.join(path, filenames[".py"]+"c"), critical = 0)
    
    # Uninstall any other items in the directory along with the
    # Python module.

    for item in requires["other items"]:

        # Use create_path as it is able to deal with platform specific
        # suffix separators.

        uninstall_res(create_path(path, item))

        if item[-3:] == ".py":

            # Uninstall any bytecode files.
            uninstall_res(create_path(path, item + "c"), critical = 0)

    for item in requires["other items"]:

        dirpath = create_path(path, item)

        dirpath, removed = os.path.split(dirpath)

        while dirpath != path:

            # The item is within a subdirectory which we must recreate
            # in the target location.

            # Check whether the directory exists but allow error
            # reporting for overwriting of files.

            if os.path.isdir(dirpath):

                uninstall_res(dirpath)

            # Find the path to the parent directory.
            dirpath, removed = os.path.split(dirpath)

    # Uninstall any generated content.

    install_dirs = {}

    for suffix, item in provides.items():

        if not item.has_key("contents"): continue

        # Retrieve the location of the resource.
        location = destinations[suffix]

        if item.has_key("directory"):

            directory = string.join(
                string.split(item["directory"] % components, "/"), os.sep
                )

            location = os.path.join(location, directory)

            install_dirs[location] = 1

        uninstall_res(
            os.path.join(location, filenames[suffix])
            )

    # Remove the support directory.

    if path != destinations[".py"]:

        uninstall_res(path)

    # Remove any other created directories.

    for location in install_dirs.keys():

        if os.path.isdir(location):

            uninstall_res(location)

    # Uninstallation for the launcher support library.

    # Uninstall the library.

    uninstall_res(
        os.path.join(destinations["library"], filenames["library"])
        )


    # Delete any images which have been installed.

    if provides.has_key("icons"):

        kde_icons_dir = destinations["icons"]

        # Uninstall the icon for each theme.

        for theme, sizes in provides["icons"]["themes"].items():

            # Construct the path to the target subdirectory associated
            # with the theme given and check whether it exists.
            theme_dir = os.path.join(kde_icons_dir, theme)

            if os.path.isdir(theme_dir):

                # Examine each size of icon provided.

                for size, images in sizes.items():

                    # Construct the path to the target subdirectory
                    # associated with the icon size given and check
                    # whether it exists.

                    size_dir = os.path.join(
                        theme_dir, size, provides["icons"]["icon type"])

                    if os.path.isdir(size_dir):

                        # Install the image using the name given in
                        # the details module.

                        for src_name, dest_name in images:

                            uninstall_res(
                                create_path(size_dir, dest_name)
                                )


# Utility function for calling kde-config

def kde_config(option):

    so, si, se = popen2.popen3("kde-config %s" % option)
    si.close()
    
    errors = se.read()
    lines = so.readlines()
    
    if errors != "":
    
        raise BuildError, "Failed to run kde-config: %s" % errors
    
    return lines

# Find the paths to use.

def resource_paths():

    lines = kde_config("--types")
    
    paths_dict = {}
    
    if errors != "":
    
        raise BuildError, "Failed to run kde-config: %s" % errors
    
    for line in lines:
    
        resource_type = "".join(line.split(" ")[:1])
        
        so, si, se = popen2.popen3("kde-config --path %s" % resource_type)
        si.close()
        errors = se.read()
        path_lines = so.readlines()
        
        if errors != "":
        
            continue
        
        paths = "".join(path_lines[:1]).strip().split(":")
        paths_dict[resource_type] = paths
    
    return paths_dict

# Prefix paths

def prefix_paths():

    paths_dict = {}
    
    root = {}
    lines = kde_config("--prefix")
    root["prefix"] = "".join(lines[:1]).strip()
    lines = kde_config("--exec-prefix")
    root["exec-prefix"] = "".join(lines[:1]).strip()
    
    paths_dict["root"] = root
    
    lines = kde_config("--localprefix")
    paths_dict["user"] = {"prefix": "".join(lines[:1]).strip()}
    
    return paths_dict

# Determine the locations of key components.

def find_components(requires):

    """components = find_components(requires)
    
    Find the components needed by the packages specified in the
    requirements dictionary. This dictionary must contain the keys
    "pyqt", "pykde" and "python".
    
    The dictionary returned contains library, include and installation
    information for Python, KDE, Qt, PyQt and PyKDE.
    
    If the operation fails, the BuildError exception is raised.
    """
    
    # Check requirements ------------------------------------------------------
    
    components = {}
    
    # Find C and C++ libraries.
    
    components["c++ libs"] = "-lstdc++ -shared-libgcc"
    
    # Find the Python library.
    
    interpreter_lib_dir = os.path.split(
        distutils.sysconfig.get_config_var("LIBP")
        )[0]
    
    # Find the Python include directory.
    interpreter_include_dir = distutils.sysconfig.get_config_var("INCLUDEPY")
    
    # Find the version of the Python interpreter and the module library
    # directories.
    
    py_version = distutils.sysconfig.get_config_var("VERSION")
    py_shared = distutils.sysconfig.get_config_var("LDLIBRARY")
    py_libdir = distutils.sysconfig.get_config_var("LIBDIR")
    py_so = distutils.sysconfig.get_config_var("SO")
    if not py_shared or not py_shared.endswith(py_so):
    
        py_shared = create_path(py_libdir, "libpython"+py_version+py_so)
        if not os.path.exists(py_shared):
        
            warning("The Python library found by distutils does not appear to be")
            warning("suitable for embedding. Ideally, if autoconf was used to")
            warning("configure Python for building, the --enable-shared option")
            warning("should have been given.")
            raise BuildError
    
    py_shared_libs = distutils.sysconfig.get_config_var("SHLIBS")
    
    components["python dir"] = py_libdir # interpreter_lib_dir
    components["python inc dir"] = interpreter_include_dir
    components["python version"] = py_version
    components["python shared lib"] = py_shared
    components["python libs"] = py_shared_libs
    
    # Find the Python library directory.
    python_libs_dir = distutils.sysconfig.get_config_var("LIBP")
    
    python_libs_subdirs = map(
        lambda x: string.strip(x), filter(
            lambda x: x != " ", string.split(
                distutils.sysconfig.get_config_var("LIBSUBDIRS"), " "
                )
            )
        )
    
    python_libs_dirs = [python_libs_dir]
    
    python_libs_dirs = python_libs_dirs + list(
        create_paths(python_libs_dir, python_libs_subdirs)
        )
    
    components["python libs dirs"] = python_libs_dirs
    
    # Construct a PyKDE configuration object to find information about how
    # Qt and KDE are configured.
    
    configuration = pykdeconfig.Configuration()
    
    # Find the Qt installation, library and include directories.
    
    try:
    
        qtdir = configuration.qt_dir
        qt_include_dir = configuration.qt_inc_dir
        qt_lib_dir = configuration.qt_lib_dir
    
    except AttributeError:
    
        warning( "Could not find a working Qt installation. Try setting the" )
        warning( "QTDIR environment variable to the appropriate value before" )
        warning( "running this script." )
        raise BuildError
    
    components["qt inc dir"] = qt_include_dir
    components["qt lib dir"] = qt_lib_dir
    
    
    # Find the KDE installation directory.
    
    try:
    
        if os.getuid() == 0:
        
            # For root users, install the component in the KDE installation
            # directory.
            
            kdedir = configuration.kdebasedir
            kde_include_dir = configuration.kdeincdir
            kde_lib_dir = configuration.kdelibdir
            install_kdedir = prefix_paths()["root"].get("prefix", None)
        
        else:
        
            # For non-root users, use the .kde directory in the user's home
            # directory to store the files.
            
            kdedir = configuration.kdebasedir
            kde_include_dir = configuration.kdeincdir
            kde_lib_dir = configuration.kdelibdir
            install_kdedir = prefix_paths()["user"].get("prefix", None)
    
    except (AttributeError, KeyError):
    
        warning(
            "Could not find a working KDE installation. Try setting the "
            "KDEDIR environment variable to the appropriate value before "
            "running this script."
            )
        raise BuildError
    
    # Create a dictionary entry which corresponds to the target location
    # for components. This will differ from the "kdedir" value if the user
    # is not root.
    
    components["kde dir"] = install_kdedir
    components["kde inc dir"] = kde_include_dir
    components["kde lib dir"] = kde_lib_dir
    components["installed kde dir"] = kdedir
    
    
    # Python module searches --------------------------------------------------
    
    # Find any other libraries required by the modules used.
    for lib, subpath in requires["python"]:
    
        if subpath is not None:
        
            # Convert the subpath specification into a platform specific
            # form.
            subpath = string.join(string.split(subpath, "/"), os.sep)
            path = os.path.join(components["python libs dirs"][0], subpath)
            
            components[lib] = (
                path, lib + distutils.sysconfig.get_config_var("SO")
                )
        
        else:
        
            # No subpath was specified so look for the module beneath the
            # main Python library directory.
            dirs = find_files(
                [("*%s" + os.extsep + "so") % lib],
                components["python libs dirs"][:1],
                return_first = 1
                )
            
            if dirs == []:
            
                warning(
                    "Failed to find the required Python module: %s" % lib
                    )
                raise BuildError
            
            # Use the first directory found and the first matching file in
            # that directory.
            path = dirs[0][0]
            file = dirs[0][1][0]
            
            components[lib] = (path, file) # create_path(path, file)
    
    return components


# Execution of actions --------------------------------------------------------

def execute_action(action):

    """execute_action(action)
    
    Execute the action specified by the user.
    """
    
    # Find the details and requirements of this module, and the resources
    # it provides.
    mod_details, requires, provides = details.find_details()
    
    # Determine the locations of key components.
    components = find_components(requires)
    
    # Actions -----------------------------------------------------------------
    
    if action == "build":
    
        filenames, destinations = find_filenames_and_destinations(
            requires, provides, components
            )
        
        return build(
            mod_details, requires, provides, components, filenames,
            destinations
            )
    
    elif action == "clean":
    
        # Remove the intermediate files created by the build process.
        
        filenames, destinations = find_filenames_and_destinations(
            requires, provides, components
            )
        
        return clean(
            mod_details, requires, provides, components, filenames,
            destinations
            )
    
    elif action == "install":
    
        # Copy the files to the appropriate destinations.
        
        filenames, destinations = find_filenames_and_destinations(
            requires, provides, components, install = True
            )
        
        return install(
            mod_details, requires, provides, components, filenames,
            destinations
            )
    
    elif action == "uninstall":
    
        # Remove the files from their installed locations.
        
        filenames, destinations = find_filenames_and_destinations(
            requires, provides, components, install = True
            )
        
        return uninstall(
            mod_details, requires, provides, components, filenames,
            destinations
            )
    
    elif action == "info":
    
        filenames, destinations = find_filenames_and_destinations(
            requires, provides, components
            )
        
        inform( "Components" )
        inform("")
        pprint.pprint (components)
        inform("")
        inform( "Filenames" )
        inform("")
        pprint.pprint (filenames)
        inform("")
        inform( "Destinations" )
        inform("")
        pprint.pprint (destinations)
        inform("")
        
        return
    
    else:
    
        raise BuildError, "Invalid action"


def import_details(details_path):

    # Make this module available in the global scope.
    global details
    
    try:
    
        details = imp.load_source("details", details_path)
    
    except (IOError, ImportError):
    
        warning( "Failed to find component details: %s" % details_path )
        raise BuildError
    
    # Return a reference to the module for the benefit of helper programs.
    return details


def read_input(argv):

    match = {}
    
    if len(argv) < 1 or len(argv) > 3:
    
        return None
    
    match["action"] = argv.pop(0)
    
    # Read the remaining arguments: there should be zero or one of these.
    
    if len(argv) == 0:
    
        pass
    
    elif len(argv) == 1:
    
        arg = argv.pop(0)
        match["path"] = arg
    
    else:
    
        return None
    
    return match


# Main program ----------------------------------------------------------------

usage = \
"""Usage: %s %s

The command must be one of the following:

  build       Build a component. The component is automatically
              configured and any necessary interface libraries
              are built.

  clean       Remove all the temporary files used when building a
              component. These include header (.h) files, object
              (.o) files, shared libraries (.so), and linking
              scripts (.la).

  install     Install a component. If this script is run by a
              normal user, the component is installed for that
              user only. If run as root, the component is
              installed with the KDE installation.

  uninstall   Uninstall a component. All files belonging to the
              component are removed from their installed locations.

  info        Provides information about a component.
"""

if __name__ == "__main__":

    syntax = '<command> [path]'
    usage = usage % (sys.argv[0], syntax)
    
    match = read_input(sys.argv[1:])
    
    if match is None:
    
        warning( usage )
        sys.exit(1)
    
    action = match["action"]
    
    # Find the directory in which this file is located.
    global build_path
    
    this_path, this_file = os.path.split(sys.argv[0])
    
    # If a build path was specified then read it, otherwise search for
    # details files in subdirectories of the directory containing this script.
    
    build_path = match.get("path", None)
    
    if build_path is None:
    
        details_files = reduce(
            lambda previous, (directory, files): previous + files,
            find_files(["details.py"], [os.path.abspath(this_path)]),
            [])
        details_paths = map(
            lambda filename: os.path.abspath(filename),
            details_files
            )
    
    else:
    
        details_paths = [create_path(build_path, "details" + os.extsep + "py")]
    
    for details_path in details_paths:
    
        # Import the details module.
        import_details(details_path)
        
        try:
        
            execute_action(action)
        
        except BuildError:
        
            warning(usage)
            break
    
    # Exit
    sys.exit()
